<?php

declare (strict_types=1);
namespace Drupal\rfc9557;

use DateTimeImmutable;
use Drupal\Component\Datetime\DateTimePlus;

/**
 * Methods for handling RFC9957 and similar date strings.
 */
final class ParseRfc9957 {

  public const DATE_PATTERN = "|^(-?\d{4,})-?(\d{2})?-?(\d{2})?|";
  public const TIME_PATTERN = "|^-?\d{4,}-\d{2}-\d{2}T(\d{2}):?(\d{2})?:?(\d{2})?|";
  public const FRACSEC_PATTERN = "|^-?\d{4,}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.(\d+)|";
  public const OFFSET_PATTERN = "#^-?\d{4,}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d+)?(\+|-)(\d{2}):(\d{2})#";
  public const Z_PATTERN = "|^-?\d{4,}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d+)?Z|";
  // https://datatracker.ietf.org/doc/html/rfc9557#section-4.1
  public const EXTENDED_PATTERN = "(?:\[(!?)([A-Za-z0-9/=_;:+.\-]+)\])";

  /**
   * Parse IXDTF (Section 4.1) into an array.
   *
   * @param string $date
   *   RFC9957 formatted date string.
   *
   * @return array
   *   Unvalidated timezone[], suffix[], and the remaining datetime.
   */
  public static function parseExtended(string $date): ?array {
    $result = [
      'datetime' => $date,
      'timezone' => NULL,
      'suffix' => [],
    ];
    $all_extended = $each_extended = [];
    if (preg_match(
      pattern: '|' . self::EXTENDED_PATTERN . '+$|',
      subject: $date,
      matches: $all_extended,
    )) {
      preg_match_all(
        pattern: '|' . self::EXTENDED_PATTERN . '|',
        subject: $all_extended[0],
        matches: $each_extended,
        flags: PREG_SET_ORDER,
      );
      $result['datetime'] = str_replace(
        subject: $date,
        search: $all_extended[0],
        replace: '',
      );
      foreach ($each_extended as $extended) {
        // Note that a time-zone is syntactically similar to a suffix-tag
        // but does not include an equals sign. This special case is only
        // available for time zone tags.
        if (!strpos(needle: '=', haystack: $extended[2])) {
          $result['timezone'] = [
            'value' => $extended[2],
            'critical' => $extended[1] == '!',
          ];
        }
        else {
          list($key, $value) = explode(separator: '=', string: $extended[2]);
          $result['suffix'][] = [
            'key' => $key,
            'value' => $value,
            'critial' => $extended[1] == '!',
          ];
        }
      }
    }

    return $result;
  }

  /**
   * Format a RFC9957 string from parts as returned by ::parseExtended()
   *
   * @param array $parts
   *  Keyed timezone[], suffix[], and the datetime.
   *
   * @return string
   *   Unvalidated RFC9957 string.
   */
  public static function formatExtended(array $parts): string {
    $result = $parts['datetime'];
    if (isset($parts['timezone']) && isset($parts['timezone']['value'])) {
      $result .= '[';
      $result .= ($parts['timezone']['critical'] ?? FALSE) ? '!' : '';
      $result .= $parts['timezone']['value'];
      $result .= ']';
    }
    if (isset($parts['suffix'])) {
      $result .= array_reduce(
        $parts['suffix'],
        fn($output, $suffix) => $output
          . '['
          . (($suffix['critical'] ?? FALSE) ? '!' : '')
          . $suffix['key']
          . '='
          . $suffix['value']
          . ']',
        ''
      );
    }
    return $result;
  }

  /**
   * Split a date string into an array of parts.
   *
   * Similar to date_parse but without validation and with different offset
   * handling for later RFC9957 validation.
   *
   * @param string $date
   *   Date in (partial) RFC3339 like format.
   *
   * @return array
   */
  public static function parseDateString(string $date): ?array {
    $date_matches = $time_matches = $fracsec_matches = $offset_matches = $extended_matches = [];
    if (!preg_match(
      pattern: self::DATE_PATTERN,
      subject: $date,
      matches: $date_matches
    )) {
      return NULL;
    }
    preg_match(
      pattern: self::TIME_PATTERN,
      subject: $date,
      matches: $time_matches
    );
    preg_match(
      pattern: self::FRACSEC_PATTERN,
      subject: $date,
      matches: $fracsec_matches,
    );

    preg_match(
      pattern: self::OFFSET_PATTERN,
      subject: $date,
      matches: $offset_matches
    );
    if (preg_match(
      pattern: self::Z_PATTERN,
      subject: $date
    )) {
      $offset_matches = ['', '-', '00', '00'];
    }

    $parsed_array = [
      'year' => $date_matches[1],
      'month' => $date_matches[2] ?? NULL,
      'day' => $date_matches[3] ?? NULL,
      'hour' => $time_matches[1] ?? NULL,
      'minute' => $time_matches[2] ?? NULL,
      'second' => $time_matches[3] ?? NULL,
      'fracsec' => $fracsec_matches[1] ?? NULL,
      'offset' => $offset_matches[1] ?? NULL,
      'offset_hour' => $offset_matches[2] ?? NULL,
      'offset_minute' => $offset_matches[3] ?? NULL,
    ];

    return $parsed_array;
  }

  /**
   * Pad a date array with default values.
   *
   * For handling Reduced Accuracy dates as full valid RFC3339 dates.
   *
   * @param array $date_array
   *   In the format from parseDateString().
   *
   * @return array
   */
  public static function padReducedAccuracy(array $date_array): array {
    // DateTimePlus would also add years.
    $date_array['month'] = $date_array['month'] ?? 1;
    $date_array['day'] = $date_array['day'] ?? 1;
    $date_array['hour'] = $date_array['hour'] ?? 0;
    $date_array['minute'] = $date_array['minute'] ?? 0;
    $date_array['second'] = $date_array['second'] ?? 0;

    return $date_array;
  }

  /**
   * Create a RFC3339 datetime string from a date array.
   *
   * Allows for floating time without offset.
   *
   * @param array $date_array
   *  In the fromat from parseDateString().
   *
   * @return string
   */
  public static function formatArrayRfc3339(array $date_array): string {
    // DateTimePlus does the padding.
    $datetime = DateTimePlus::arrayToISO($date_array);
    if ($date_array['offset_minute']) {
      $datetime .= '+' . $date_array['offset_hour'] . ':' . $date_array['offset_minute'];
    }
    return $datetime;
  }

  /**
   * Validate a RFC3339 like datetime string.
   *
   * @param string $datetime
   *   Datetime to be validated.
   *
   * @return bool
   *   True if it is valid.
   */
  public static function validateDateTime(string $datetime, bool $strict): bool {
    $date_parts = self::parseDateString($datetime);
    $format = \DateTime::RFC3339;
    // Can have fractions of seconds.
    if (!is_null($date_parts['fracsec'])) {
      $format = \DateTime::RFC3339_EXTENDED;
    }
    // If not strict allow years outside 0000 - 9999.
    if (!$strict) {
      $format = str_replace('Y', 'X', $format);
    }
    // Allowed to validate without offset.
    if (is_null($date_parts['offset'])) {
      $format = substr($format, 0, -1);
    }
    try {
      $date_parsed = date_parse_from_format($format, $datetime);
      return empty($date_parsed['warnings']) && empty($date_parsed['errors']);
    }
    catch (\ValueError $e) {
      return FALSE;
    }
  }

  /**
   * Validate a RFC9957 extended information time zone.
   *
   * @param string $timezone
   *   The time zone to validate.
   *
   * @return bool
   *   True if it is valid.
   */
  public static function validateTimeZone(string $timezone): bool {
    // DateTimeZone is much more relaxed about what will create an object.

    // An offset is a valid time zone.
    if (self::isOffset($timezone)) {
      return TRUE;
    }
    return in_array($timezone, timezone_identifiers_list());
  }

  /**
   * Test if a string is an offset.
   *
   * Does not validate if it is valid offset.
   *
   * @param string $timezone
   *
   * @return bool
   */
  public static function isOffset(string $timezone): bool {
    return (bool) preg_match(',^[+\-]\d{2}:\d{2}$,', $timezone);
  }

  /**
   * Check for consistentency of RFC3339 time-offset and RFC9957 time zone.
   *
   * @param string $datetime
   *   Datetime with offset.
   * @param string $timezone
   *   Extended information time zone.
   *
   * @return bool
   *   True if the time-offset and time zone combination are valid.
   */
  public static function validateOffsetTimeZone(string $datetime, string $timezone): bool {
    $date_parts = self::parseDateString($datetime);
    if (self::isOffset($timezone)) {
      // If the timezone is also an offset they should be the same.
      return ($date_parts['offset'] . $date_parts['offset_hour'] . ':' .$date_parts['offset_minute']) == $timezone;
    }

    $format = \DateTime::RFC3339;
    if (!is_null($date_parts['fracsec'])) {
      $format = \DateTime::RFC3339_EXTENDED;
    }
    $datetime_offset = DateTimeImmutable::createFromFormat($format, $datetime);

    $date_parts['offset_hour'] = $date_parts['offset_minute'] = $date_parts['offset'] = NULL;
    $datetimezone = DateTimeImmutable::createFromFormat(substr($format, 0, -1), self::formatArrayRfc3339($date_parts), new \DateTimeZone($timezone));

    if ($datetime_offset->getTimestamp() != $datetimezone->getTimestamp()) {
      return FALSE;
    }

    return TRUE;
  }

}
