<?php

declare (strict_types=1);
namespace Drupal\rfc9557\Plugin\DataType;

use DateTimeZone;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\Attribute\DataType;
use Drupal\Core\TypedData\Plugin\DataType\StringData;
use Drupal\Core\TypedData\Type\DateTimeInterface;
use Drupal\rfc9557\ParseRfc9957;
use Drupal\rfc9557\Plugin\Validation\Constraint\Rfc9557;

/**
 * Date time typed data for RFC9557 extended dates.
 */
#[DataType(
  id: "datetime_rfc9557",
  label: new TranslatableMarkup("DateTime RFC 9557"),
  description: new TranslatableMarkup("Datetime with optional time zone."),
  constraints: [
    'Rfc9557' => ['mode' => Rfc9557::MODE_STRICT],
  ],
)]
class DateTimeRfc9557 extends StringData implements DateTimeInterface {

  /**
   * {@inheritdoc}
   */
  public function getDateTime(): ?DrupalDateTime {
    if ($this->value) {
      $parts = ParseRfc9957::parseExtended($this->value);
      $parsed = ParseRfc9957::parseDateString($parts['datetime']);
      $iso = ParseRfc9957::formatArrayRfc3339(ParseRfc9957::padReducedAccuracy($parsed));
      // @todo Offset vs. Timezone handling.
      $date_time = new DrupalDateTime($iso);
      if (is_array($parts['timezone'])) {
        $date_time->setTimezone(new DateTimeZone($parts['timezone']['value']));
      }
      return $date_time;
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function setDateTime(DrupalDateTime $dateTime, $notify = TRUE): void {
    // If updating an existing value, maintain format as much as makes sense.
    // If maintaining accuracy makes sense is questionable.
    // @todo perhaps additional methods to manipulate specific parts more
    // explicitly?
    $parts = [];
    $z = FALSE;
    if ($this->value) {
      $parts = ParseRfc9957::parseExtended($this->value);
      $parsed = ParseRfc9957::parseDateString($parts['datetime']);
      $iso = ParseRfc9957::formatArrayRfc3339(ParseRfc9957::padReducedAccuracy($parsed));
      $z = $parsed['offset'] === '-' && $parsed['offset_hour'] === '00' && $parsed['offset_minute'] === '00';
    }
    if (in_array($dateTime->getTimezone()->getName(), DateTimeZone::listIdentifiers())) {
      $parts['timezone']['value'] = $dateTime->getTimezone()->getName();
    }
    if ($z) {
      $dateTime->setTimezone(new DateTimeZone('UTC'));
      // Replacement p will treat '-00:00' but also '+00:00' as Z so can't be
      // used more generally.
      // https://www.rfc-editor.org/rfc/rfc9557#name-update-to-rfc-3339
      $parts['datetime'] = $dateTime->format('Y-m-d\TH:i:sp');
    }
    else {
      $parts['datetime'] = $dateTime->format('c');
    }
    $this->setValue(ParseRfc9957::formatExtended($parts), $notify);
  }

}
