<?php

declare (strict_types=1);
namespace Drupal\rfc9557\Plugin\Validation\Constraint;

use Drupal\rfc9557\ParseRfc9957;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;

/**
 * Validates the Rfc9557 constraint.
 */
class Rfc9557Validator extends ConstraintValidator {

  /**
   * {@inheritdoc}
   */
  public function validate(mixed $value, Constraint $constraint): void {
    if (!$constraint instanceof Rfc9557) {
      throw new \UnexpectedTypeException($constraint, __NAMESPACE__ . '\Rfc9557');
    }

    if (NULL === $value) {
      return;
    }

    $parts = ParseRfc9957::parseExtended($value);
    $datetime = $parts['datetime'];
    $date_parts = ParseRfc9957::parseDateString($datetime);
    if ($constraint->mode == Rfc9557::MODE_REDUCED_ACCURACY) {
      $date_parts = ParseRfc9957::padReducedAccuracy($date_parts);
      $datetime = ParseRfc9957::formatArrayRfc3339($date_parts);
    }
    if (!ParseRfc9957::validateDateTime($datetime, $constraint->mode == Rfc9557::MODE_STRICT)) {
      $this->context->buildViolation($constraint->invalidDatetimeMessage)
        ->setCode(Rfc9557::INVALID_DATETIME)
        ->addViolation();

      return;
    }
    if ($constraint->mode == Rfc9557::MODE_STRICT
      && empty($date_parts['offset'])
    ) {
      $this->context->buildViolation($constraint->invalidDatetimeMessage)
        ->setCode(Rfc9557::INVALID_DATETIME)
        ->addViolation();

      return;
    }
    if (!is_null($parts['timezone']) &&
      !ParseRfc9957::validateTimeZone($parts['timezone']['value'])
    ) {
      $this->context->buildViolation($constraint->invalidTimeZoneMessage)
        ->setCode(Rfc9557::INVALID_TIMEZONE)
        ->addViolation();

      return;
    }
    // An offset in a extended information time zone needs to be compared.
    // Otherwise if the time zone is marked as critical.
    // Inconsistency can also be resolved.
    // https://datatracker.ietf.org/doc/html/rfc9557#section-3.4
    if ($constraint->mode == Rfc9557::MODE_STRICT &&
      !($date_parts['offset_hour'] == 0 && $date_parts['offset'] == '-') &&
      !is_null($parts['timezone']) &&
      ($parts['timezone']['critical'] || ParseRfc9957::isOffset($parts['timezone']['value'])) &&
      !ParseRfc9957::validateOffsetTimeZone($datetime, $parts['timezone']['value'])
    ) {
      $this->context->buildViolation($constraint->offsetInconsistentMessage)
        ->setCode(Rfc9557::OFFSET_INCONSISTENT)
        ->addViolation();

      return;
    }
  }

}
